"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const shuffle = require("lodash.shuffle");
const errors_1 = require("./errors");
const peer_1 = require("./peer");
exports.EVENT_CLOSE_OUTBOUND = peer_1.EVENT_CLOSE_OUTBOUND;
exports.EVENT_CONNECT_ABORT_OUTBOUND = peer_1.EVENT_CONNECT_ABORT_OUTBOUND;
exports.EVENT_CONNECT_OUTBOUND = peer_1.EVENT_CONNECT_OUTBOUND;
exports.EVENT_FAILED_PEER_INFO_UPDATE = peer_1.EVENT_FAILED_PEER_INFO_UPDATE;
exports.EVENT_INBOUND_SOCKET_ERROR = peer_1.EVENT_INBOUND_SOCKET_ERROR;
exports.EVENT_MESSAGE_RECEIVED = peer_1.EVENT_MESSAGE_RECEIVED;
exports.EVENT_OUTBOUND_SOCKET_ERROR = peer_1.EVENT_OUTBOUND_SOCKET_ERROR;
exports.EVENT_REQUEST_RECEIVED = peer_1.EVENT_REQUEST_RECEIVED;
exports.EVENT_UPDATED_PEER_INFO = peer_1.EVENT_UPDATED_PEER_INFO;
const peer_discovery_1 = require("./peer_discovery");
exports.EVENT_FAILED_TO_PUSH_NODE_INFO = 'failedToPushNodeInfo';
exports.EVENT_DISCOVERED_PEER = 'discoveredPeer';
exports.EVENT_FAILED_TO_FETCH_PEER_INFO = 'failedToFetchPeerInfo';
exports.MAX_PEER_LIST_BATCH_SIZE = 100;
exports.MAX_PEER_DISCOVERY_PROBE_SAMPLE_SIZE = 100;
const selectRandomPeerSample = (peerList, count) => shuffle(peerList).slice(0, count);
class PeerPool extends events_1.EventEmitter {
    constructor(peerPoolConfig) {
        super();
        this._peerMap = new Map();
        this._peerPoolConfig = peerPoolConfig;
        this._peerSelectForSendRequest = peerPoolConfig.peerSelectionForSendRequest;
        this._peerSelectForConnection = peerPoolConfig.peerSelectionForConnection;
        this._handlePeerRPC = (request) => {
            this.emit(peer_1.EVENT_REQUEST_RECEIVED, request);
        };
        this._handlePeerMessage = (message) => {
            this.emit(peer_1.EVENT_MESSAGE_RECEIVED, message);
        };
        this._handlePeerConnect = async (peerInfo) => {
            this.emit(peer_1.EVENT_CONNECT_OUTBOUND, peerInfo);
            const peerId = peer_1.constructPeerIdFromPeerInfo(peerInfo);
            const peer = this.getPeer(peerId);
            if (!peer) {
                this.emit(exports.EVENT_FAILED_TO_FETCH_PEER_INFO, new errors_1.RequestFailError('Failed to fetch peer status because the relevant peer could not be found'));
                return;
            }
            let detailedPeerInfo;
            try {
                detailedPeerInfo = await peer.fetchStatus();
            }
            catch (error) {
                this.emit(exports.EVENT_FAILED_TO_FETCH_PEER_INFO, error);
                return;
            }
            this.emit(exports.EVENT_DISCOVERED_PEER, detailedPeerInfo);
        };
        this._handlePeerConnectAbort = (peerInfo) => {
            this.emit(peer_1.EVENT_CONNECT_ABORT_OUTBOUND, peerInfo);
        };
        this._handlePeerClose = (closePacket) => {
            const peerId = peer_1.constructPeerIdFromPeerInfo(closePacket.peerInfo);
            this.removePeer(peerId);
            this.emit(peer_1.EVENT_CLOSE_OUTBOUND, closePacket);
        };
        this._handlePeerOutboundSocketError = (error) => {
            this.emit(peer_1.EVENT_OUTBOUND_SOCKET_ERROR, error);
        };
        this._handlePeerInboundSocketError = (error) => {
            this.emit(peer_1.EVENT_INBOUND_SOCKET_ERROR, error);
        };
        this._handlePeerInfoUpdate = (peerInfo) => {
            this.emit(peer_1.EVENT_UPDATED_PEER_INFO, peerInfo);
        };
        this._handleFailedPeerInfoUpdate = (error) => {
            this.emit(peer_1.EVENT_FAILED_PEER_INFO_UPDATE, error);
        };
    }
    applyNodeInfo(nodeInfo) {
        this._nodeInfo = nodeInfo;
        const peerList = this.getAllPeers();
        peerList.forEach(peer => {
            this._applyNodeInfoOnPeer(peer, nodeInfo);
        });
    }
    get nodeInfo() {
        return this._nodeInfo;
    }
    selectPeers(numOfPeers) {
        const listOfPeerInfo = [...this._peerMap.values()].map((peer) => peer.peerInfo);
        const selectedPeers = this._peerSelectForSendRequest(listOfPeerInfo, this._nodeInfo, numOfPeers);
        return selectedPeers;
    }
    async requestFromPeer(packet) {
        const selectedPeer = this.selectPeers(1);
        if (selectedPeer.length <= 0) {
            throw new errors_1.RequestFailError('Request failed due to no peers found in peer selection');
        }
        const selectedPeerId = peer_1.constructPeerIdFromPeerInfo(selectedPeer[0]);
        const peer = this._peerMap.get(selectedPeerId);
        if (!peer) {
            throw new errors_1.RequestFailError(`No such Peer exist in PeerPool with the selected peer with Id: ${selectedPeerId}`);
        }
        const response = await peer.request(packet);
        return response;
    }
    sendToPeers(message) {
        const selectedPeers = this.selectPeers();
        selectedPeers.forEach((peerInfo) => {
            const selectedPeerId = peer_1.constructPeerIdFromPeerInfo(peerInfo);
            const peer = this._peerMap.get(selectedPeerId);
            if (peer) {
                peer.send(message);
            }
        });
    }
    async fetchStatusAndCreatePeers(seedPeers, nodeInfo, peerConfig) {
        const listOfPeerInfos = await Promise.all(seedPeers.map(async (seedPeer) => {
            try {
                const seedFetchStatusResponse = await peer_1.connectAndFetchPeerInfo(seedPeer, nodeInfo, peerConfig);
                const peerId = peer_1.constructPeerIdFromPeerInfo(seedFetchStatusResponse.peerInfo);
                this.addOutboundPeer(peerId, seedFetchStatusResponse.peerInfo, seedFetchStatusResponse.socket);
                return seedFetchStatusResponse.peerInfo;
            }
            catch (error) {
                this.emit(exports.EVENT_FAILED_TO_FETCH_PEER_INFO, error);
                return undefined;
            }
        }));
        const filteredListOfPeers = listOfPeerInfos.filter(peerInfo => peerInfo !== undefined);
        return filteredListOfPeers;
    }
    async runDiscovery(knownPeers, blacklist) {
        const peersForDiscovery = knownPeers.map(peerInfo => {
            const peerId = peer_1.constructPeerIdFromPeerInfo(peerInfo);
            const existingPeer = this.getPeer(peerId);
            if (existingPeer) {
                existingPeer.updatePeerInfo(peerInfo);
                return existingPeer;
            }
            return this.addPeer(peerInfo);
        });
        const peerSampleToProbe = selectRandomPeerSample(peersForDiscovery, exports.MAX_PEER_DISCOVERY_PROBE_SAMPLE_SIZE);
        const discoveredPeers = await peer_discovery_1.discoverPeers(peerSampleToProbe, {
            blacklist: blacklist.map(peer => peer.ipAddress),
        });
        discoveredPeers.forEach((peerInfo) => {
            const peerId = peer_1.constructPeerIdFromPeerInfo(peerInfo);
            const existingPeer = this.getPeer(peerId);
            if (existingPeer) {
                existingPeer.updatePeerInfo(peerInfo);
            }
        });
        return discoveredPeers;
    }
    selectPeersAndConnect(peers) {
        const peersToConnect = this._peerSelectForConnection(peers);
        peersToConnect.forEach((peerInfo) => {
            const peerId = peer_1.constructPeerIdFromPeerInfo(peerInfo);
            const existingPeer = this.getPeer(peerId);
            if (!existingPeer) {
                return this.addPeer(peerInfo);
            }
            existingPeer.updatePeerInfo(peerInfo);
            return existingPeer;
        });
        return peersToConnect;
    }
    addPeer(peerInfo, inboundSocket) {
        const peerConfig = {
            connectTimeout: this._peerPoolConfig.connectTimeout,
            ackTimeout: this._peerPoolConfig.ackTimeout,
        };
        const peer = new peer_1.Peer(peerInfo, peerConfig, { inbound: inboundSocket });
        if (this._peerMap.has(peer.id)) {
            throw new Error(`Peer ${peer.id} was already in the peer pool`);
        }
        this._peerMap.set(peer.id, peer);
        this._bindHandlersToPeer(peer);
        if (this._nodeInfo) {
            this._applyNodeInfoOnPeer(peer, this._nodeInfo);
        }
        peer.connect();
        return peer;
    }
    addDiscoveredPeer(detailedPeerInfo, inboundSocket) {
        const peerConfig = {
            connectTimeout: this._peerPoolConfig.connectTimeout,
            ackTimeout: this._peerPoolConfig.ackTimeout,
        };
        const peer = new peer_1.Peer(detailedPeerInfo, peerConfig, {
            inbound: inboundSocket,
        });
        this._peerMap.set(peer.id, peer);
        this._bindHandlersToPeer(peer);
        if (this._nodeInfo) {
            this._applyNodeInfoOnPeer(peer, this._nodeInfo);
        }
        peer.updatePeerInfo(detailedPeerInfo);
        peer.connect();
        return peer;
    }
    addInboundPeer(peerId, peerInfo, socket) {
        const existingPeer = this.getPeer(peerId);
        if (existingPeer) {
            existingPeer.updatePeerInfo(peerInfo);
            if (existingPeer.state.inbound === peer_1.ConnectionState.DISCONNECTED) {
                existingPeer.inboundSocket = socket;
                return false;
            }
            return false;
        }
        this.addPeer(peerInfo, socket);
        return true;
    }
    addOutboundPeer(peerId, peerInfo, socket) {
        const existingPeer = this.getPeer(peerId);
        if (existingPeer) {
            existingPeer.updatePeerInfo(peerInfo);
            if (existingPeer.state.outbound === peer_1.ConnectionState.DISCONNECTED) {
                existingPeer.outboundSocket = socket;
                return false;
            }
            return false;
        }
        const peerConfig = {
            connectTimeout: this._peerPoolConfig.connectTimeout,
            ackTimeout: this._peerPoolConfig.ackTimeout,
        };
        const peer = new peer_1.Peer(peerInfo, peerConfig, { outbound: socket });
        this._peerMap.set(peer.id, peer);
        this._bindHandlersToPeer(peer);
        if (this._nodeInfo) {
            this._applyNodeInfoOnPeer(peer, this._nodeInfo);
        }
        return true;
    }
    removeAllPeers() {
        this._peerMap.forEach((peer) => {
            this.removePeer(peer.id);
        });
    }
    getAllPeerInfos() {
        return this.getAllPeers().map(peer => peer.peerInfo);
    }
    getAllPeers() {
        return [...this._peerMap.values()];
    }
    getPeer(peerId) {
        return this._peerMap.get(peerId);
    }
    hasPeer(peerId) {
        return this._peerMap.has(peerId);
    }
    removePeer(peerId) {
        const peer = this._peerMap.get(peerId);
        if (peer) {
            peer.disconnect();
            this._unbindHandlersFromPeer(peer);
        }
        return this._peerMap.delete(peerId);
    }
    _applyNodeInfoOnPeer(peer, nodeInfo) {
        (async () => {
            try {
                await peer.applyNodeInfo(nodeInfo);
            }
            catch (error) {
                this.emit(exports.EVENT_FAILED_TO_PUSH_NODE_INFO, error);
            }
        })();
    }
    _bindHandlersToPeer(peer) {
        peer.on(peer_1.EVENT_REQUEST_RECEIVED, this._handlePeerRPC);
        peer.on(peer_1.EVENT_MESSAGE_RECEIVED, this._handlePeerMessage);
        peer.on(peer_1.EVENT_CONNECT_OUTBOUND, this._handlePeerConnect);
        peer.on(peer_1.EVENT_CONNECT_ABORT_OUTBOUND, this._handlePeerConnectAbort);
        peer.on(peer_1.EVENT_CLOSE_OUTBOUND, this._handlePeerClose);
        peer.on(peer_1.EVENT_OUTBOUND_SOCKET_ERROR, this._handlePeerOutboundSocketError);
        peer.on(peer_1.EVENT_INBOUND_SOCKET_ERROR, this._handlePeerInboundSocketError);
        peer.on(peer_1.EVENT_UPDATED_PEER_INFO, this._handlePeerInfoUpdate);
        peer.on(peer_1.EVENT_FAILED_PEER_INFO_UPDATE, this._handleFailedPeerInfoUpdate);
    }
    _unbindHandlersFromPeer(peer) {
        peer.removeListener(peer_1.EVENT_REQUEST_RECEIVED, this._handlePeerRPC);
        peer.removeListener(peer_1.EVENT_MESSAGE_RECEIVED, this._handlePeerMessage);
        peer.removeListener(peer_1.EVENT_CONNECT_OUTBOUND, this._handlePeerConnect);
        peer.removeListener(peer_1.EVENT_CONNECT_ABORT_OUTBOUND, this._handlePeerConnectAbort);
        peer.removeListener(peer_1.EVENT_CLOSE_OUTBOUND, this._handlePeerClose);
        peer.removeListener(peer_1.EVENT_UPDATED_PEER_INFO, this._handlePeerInfoUpdate);
        peer.removeListener(peer_1.EVENT_FAILED_PEER_INFO_UPDATE, this._handleFailedPeerInfoUpdate);
    }
}
exports.PeerPool = PeerPool;
//# sourceMappingURL=peer_pool.js.map